<?php
namespace AkTest\App;

use Ak\App\Application;
use Ak\App\ApplicationInterface;
use Ak\Test\AbstractTestCase;
use PHPUnit\Framework\MockObject\MockObject;
use Psr\Container\ContainerExceptionInterface;
use Psr\Container\ContainerInterface;
use Psr\Container\NotFoundExceptionInterface;
use Psr\Http\Server\MiddlewareInterface;
use Psr\Http\Server\RequestHandlerInterface;
use Laminas\Diactoros\Response\TextResponse;
use Laminas\Stratigility\Middleware\RequestHandlerMiddleware;
use ReflectionException;

/**
 * Class ApplicationTest
 *
 * @package AkTest\App
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/autokreso/app
 */
final class ApplicationTest extends AbstractTestCase
{
    private array $files = [
        'apps/blank/var/config.php',
        'apps/blank/var',
        'apps/fs-app/var/config.php',
        'apps/fs-app/var',
        'var',
    ];

    protected function setUp(): void
    {
        parent::setUp();
        $this->deleteFiles();
    }

    protected function tearDown(): void
    {
        parent::tearDown();
        $this->deleteFiles();
    }

    private function deleteFiles()
    {
        chdir(__DIR__);

        foreach ($this->files as $file) {
            if (file_exists($file)) {
                if (is_dir($file)) {
                    rmdir($file);
                } else {
                    unlink($file);
                }
            }
        }
    }

    public function testCanDetectDevelopmentMode()
    {
        $application = new Application(__DIR__);
        $this->assertFalse($application->isDevelopment());

        putenv('APPLICATION_ENV=development');

        $application = new Application(__DIR__);
        $this->assertTrue($application->isDevelopment());

        putenv('APPLICATION_ENV=');

        $application = new Application(__DIR__);
        $this->assertFalse($application->isDevelopment());
    }

    public function testWillCacheTheConfig()
    {
        $application = new Application(__DIR__ . '/apps/blank');
        $config = $application->getConfig();

        $this->assertFileExists(__DIR__ . '/apps/blank/var/config.php');

        /** @noinspection PhpIncludeInspection */
        $cached = include __DIR__ . '/apps/blank/var/config.php';

        $this->assertEquals($config, $cached);
    }

    public function testWillLoadTheProviders()
    {
        $application = new Application(__DIR__ . '/apps/fs-app');
        $config = $application->getConfig();

        $this->assertArrayHasKey('file_loader', $config['dependencies']['aliases']);
    }

    public function testCanCreateContainer()
    {
        $application = new Application(__DIR__ . '/apps/fs-app');
        $container = $application->getContainer();

        $this->assertTrue($container->has('file_loader'));
    }

    /**
     * @return void
     * @throws ContainerExceptionInterface
     * @throws NotFoundExceptionInterface
     */
    public function testContainerIsRegisteredUnderPsrInterface()
    {
        $application = new Application(__DIR__ . '/apps/blank');
        $container = $application->getContainer();

        $this->assertTrue($container->has(ContainerInterface::class));
        $this->assertSame($container, $container->get(ContainerInterface::class));
    }

    public function testApplicationRegistersItselfInTheContainer()
    {
        $application = new Application(__DIR__ . '/apps/blank');
        $container = $application->getContainer();

        $this->assertContainerHasService($container, [
            'app',
            ApplicationInterface::class,
        ], ApplicationInterface::class);
    }

    /**
     * @throws ReflectionException
     */
    public function testCanCreateMiddlewarePipe()
    {
        $application = new Application(__DIR__ . '/apps/blank');
        $pipe = $application->getPipe();

        $class = new \ReflectionClass($pipe);
        $pipeline = $class->getProperty('pipeline');
        $pipeline->setAccessible(true);

        $middlewares = $pipeline->getValue($pipe);

        $this->assertCount(0, $middlewares);
    }

    /**
     * @throws ReflectionException
     */
    public function testCanPipeMiddlewareDirectly()
    {
        /** @var MiddlewareInterface|MockObject $middleware */
        $middleware = $this->createMock(MiddlewareInterface::class);

        $application = new Application(__DIR__ . '/apps/blank');
        $application->pipe($middleware);
        $pipe = $application->getPipe();

        $class = new \ReflectionClass($pipe);
        $pipeline = $class->getProperty('pipeline');
        $pipeline->setAccessible(true);

        $middlewares = $pipeline->getValue($pipe);

        $this->assertCount(1, $middlewares);
        $this->assertSame($middleware, $middlewares[0]);
    }

    /**
     * @runInSeparateProcess
     */
    public function testRunnerCanRunTheApp()
    {
        $response = new TextResponse('it works');
        /** @var RequestHandlerMiddleware|MockObject $myApp */
        $myApp = $this->createMock(RequestHandlerInterface::class);
        $myApp->expects($this->once())
            ->method('handle')
            ->willReturn($response);

        $application = new Application(__DIR__ . '/apps/blank');
        $application->pipe(new RequestHandlerMiddleware($myApp));

        ob_start();
        $application->run();
        $output = ob_get_clean();

        $this->assertEquals('it works', $output);
    }

    public function testCanCreateErrorResponse()
    {
        $application = new Application(__DIR__ . '/apps/blank');
        $error = new \Exception('error message');

        $response = $application->createErrorResponse($error);

        $this->assertEquals(500, $response->getStatusCode());
    }

    public function testErrorResponseWillContainerErrorDataInDevelopmentMode()
    {
        putenv('APPLICATION_ENV=development');
        $application = new Application(__DIR__ . '/apps/blank');
        $error = new \Exception('error message', 502);

        $response = $application->createErrorResponse($error);
        $response->getBody()->rewind();

        $this->assertEquals(502, $response->getStatusCode());
        $this->assertStringContainsString('error message', $response->getBody()->getContents());

        putenv('APPLICATION_ENV=');
    }
}
